/*
 *
 * LoopS: Structurally Dissolvable Self-Reproducing Loop Simulator
 *
 *             Ver. 1.02 (slightly bug-fixed version)
 *
 *              Copyright (C) 1998 by SAYAMA Hiroki
 *                    the University of Tokyo
 *
 *
 *         If you find questions or bugs in this program,
 *                  please inform to the author.
 *                  (sayama@is.s.u-tokyo.ac.jp)
 *
 */


/* Definition of constants */

#define CYCLIC 0
#define FREE   1
#define OFF    0
#define ON     1

#define MARGIN   20
#define TIMEFONT "-adobe-helvetica-bold-r-normal--14-140-75-75-p-82-iso8859-1"

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <string.h>


/* Initialization of window and colors */

typedef struct ground {
  Display *disp;
  Window  window;
  GC      gc;
  int     w, h, boundary, pixel;
  int     screen, pause, time;
  unsigned long color[9];
} GROUND;

void open_window(g)
     GROUND *g;
{
  XSizeHints hints;
  XEvent eve;

  g->disp = XOpenDisplay(NULL);
  if (g->time == OFF)
    g->window = XCreateSimpleWindow(g->disp, DefaultRootWindow(g->disp), 0, 0,
				    g->w * g->pixel, g->h * g->pixel,
				    2,
				    WhitePixel(g->disp, 0),
				    BlackPixel(g->disp, 0));
  else 
    g->window = XCreateSimpleWindow(g->disp, DefaultRootWindow(g->disp), 0, 0,
				    g->w * g->pixel, g->h * g->pixel + MARGIN,
				    2,
				    WhitePixel(g->disp, 0),
				    BlackPixel(g->disp, 0));

  hints.flags = USPosition | USSize;
  XSetWMNormalHints(g->disp, g->window, &hints);
  XStoreName(g->disp, g->window, "LoopS");
  XMapWindow(g->disp, g->window);
  XSelectInput(g->disp, g->window,
	       ButtonPressMask | ExposureMask | KeyPressMask |KeyReleaseMask);

  g->gc = XCreateGC(g->disp, g->window, 0, 0);
  XSetFont(g->disp, g->gc, XLoadFont(g->disp, TIMEFONT));

  XFlush(g->disp);

  while (1)
    if (XCheckWindowEvent(g->disp, g->window, ExposureMask, &eve) == True)
      break;
}

unsigned long makecolor(g, R, G, B)
     GROUND *g;
     unsigned short R, G, B;
{
  XColor xc;

  xc.red   = R;
  xc.green = G;
  xc.blue  = B;
  XAllocColor(g->disp, DefaultColormap(g->disp, 0), &xc);
  return xc.pixel;
}

void set_color(g)
     GROUND *g;
{
  g->color[0] = makecolor(g,     0,     0,     0);
  g->color[1] = makecolor(g,     0,     0, 65535);
  g->color[2] = makecolor(g, 65535,     0,     0);
  g->color[3] = makecolor(g, 65535,     0, 65535);
  g->color[4] = makecolor(g,     0, 65535,     0);
  g->color[5] = makecolor(g,     0, 65535, 65535);
  g->color[6] = makecolor(g, 65535, 65535,     0);
  g->color[7] = makecolor(g, 65535, 65535, 65535);
  g->color[8] = makecolor(g, 32768, 32768, 32768);
}


/* Definition of transition rules */

void load_rule(rule, rule_file, symmetry)
     int rule[9][9][9][9][9];
     char *rule_file;
     int symmetry;
{
  int  a, b, c, d, e, f;
  char C, T, R, B, L, I;

  FILE *fp;
  char buf[256];

  for (a = 0; a <= 8; a ++)
    for (b = 0; b <= 8; b ++)
      for (c = 0; c <= 8; c ++)
	for (d = 0; d <= 8; d ++)
	  for (e = 0; e <= 8; e ++) rule[a][b][c][d][e] = -1;

  if ((fp = fopen (rule_file, "r")) == NULL) {
    fprintf(stderr, "Rule file \"%s\" cannot be opened.\n", rule_file);
    exit();
  }

  while (1) {
    if (fgets(buf, 256, fp) == NULL) break;

    sscanf(buf, "%c%c%c%c%c%c", &C, &T, &R, &B, &L, &I);
    a = (int) (C - '0');
    b = (int) (T - '0');
    c = (int) (R - '0');
    d = (int) (B - '0');
    e = (int) (L - '0');
    f = (int) (I - '0');

    if ((a < 0) || (a > 8) ||
	(b < 0) || (b > 8) ||
	(c < 0) || (c > 8) ||
	(d < 0) || (d > 8) ||
	(e < 0) || (e > 8) ||
	(f < 0) || (f > 8)) {
      fprintf(stderr, "Illegal state appeared in \"%s\".\n", rule_file);
      exit();
    }

    rule[a][b][c][d][e] = f;

    if (symmetry == ON) {
      rule[a][c][d][e][b] = f;
      rule[a][d][e][b][c] = f;
      rule[a][e][b][c][d] = f;
    }
  }

  fclose(fp);
}

int in_sheath(b, c, d, e)
     int b, c, d, e;
{
  int f = 0;

  if ((b == 1) || (b == 2) || (b == 4) || (b == 6) || (b == 7)) f ++;
  if ((c == 1) || (c == 2) || (c == 4) || (c == 6) || (c == 7)) f ++;
  if ((d == 1) || (d == 2) || (d == 4) || (d == 6) || (d == 7)) f ++;
  if ((e == 1) || (e == 2) || (e == 4) || (e == 6) || (e == 7)) f ++;

  if (f >= 2) return True;
  else return False;
}

void set_undefined_rule(rule, extension, sd)
     int rule[9][9][9][9][9], extension, sd;
{
  int a, b, c, d, e;


  /* Extension of rules */

  if (extension == ON) {
    for (a = 0; a <= 7; a ++)
      for (b = 0; b <= 7; b ++)
	for (c = 0; c <= 7; c ++)
	  for (d = 0; d <= 7; d ++)
	    for (e = 0; e <= 7; e ++)
	      if (rule[a][b][c][d][e] == -1)
		switch (a) {

		case 0:
		  if (((b == 1) || (c == 1) || (d == 1) || (e == 1)) &&
		      (in_sheath(b, c, d, e) == True))
		    rule[a][b][c][d][e] = 1;
		  else
		    rule[a][b][c][d][e] = 0;
		  break;

		case 1:
		  if (((b == 7) || (c == 7) || (d == 7) || (e == 7)) &&
		      (in_sheath(b, c, d, e) == True))
		    rule[a][b][c][d][e] = 7;
		  else if (((b == 6) || (c == 6) || (d == 6) || (e == 6)) &&
			   (in_sheath(b, c, d, e) == True))
		    rule[a][b][c][d][e] = 6;
		  else if (((b == 4) || (c == 4) || (d == 4) || (e == 4)) &&
			   (in_sheath(b, c, d, e) == True))
		    rule[a][b][c][d][e] = 4;
		  break;

		case 2:
		  if ((b == 3) || (c == 3) || (d == 3) || (e == 3))
		    rule[a][b][c][d][e] = 1;
		  else if ((b == 2) || (c == 2) || (d == 2) || (e == 2))
		    rule[a][b][c][d][e] = 2;
		  break;

		case 4:
		case 6:
		case 7:
		  if (((b == 0) || (c == 0) || (d == 0) || (e == 0)) &&
		      (in_sheath(b, c, d, e) == True))
		    rule[a][b][c][d][e] = 0;
		  break;
		}
  }


  /* Definition of rules which concern the state 8 */

  if (sd == ON) {

    for (a = 0; a <= 8; a ++)
      for (b = 0; b <= 8; b ++)
	for (c = 0; c <= 8; c ++)
	  for (d = 0; d <= 8; d ++)
	    for (e = 0; e <= 8; e ++)
	      if (rule[a][b][c][d][e] == -1) {

		if (a == 8) rule[a][b][c][d][e] = 0;

		else if ((b == 8) || (c == 8) || (d == 8) || (e == 8))
		  switch(a) {

		  case 0:
		  case 1:
		    if (((b >= 2) && (b <= 7)) || ((c >= 2) && (c <= 7)) ||
			((d >= 2) && (d <= 7)) || ((e >= 2) && (e <= 7)))
		      rule[a][b][c][d][e] = 8;
		    else rule[a][b][c][d][e] = a;
		    break;

		  case 2:
		  case 3:
		  case 5:
		    rule[a][b][c][d][e] = 0;
		    break;

		  case 4:
		  case 6:
		  case 7:
		    rule[a][b][c][d][e] = 1;
		    break;
		  }
	      }

    rule[1][1][1][5][2] = 8;
    rule[1][1][5][2][1] = 8;
    rule[1][5][2][1][1] = 8;
    rule[1][2][1][1][5] = 8;
  }


  /* Clearance of undefined rules */

  for (a = 0; a <= 8; a ++)
    for (b = 0; b <= 8; b ++)
      for (c = 0; c <= 8; c ++)
	for (d = 0; d <= 8; d ++)
	  for (e = 0; e <= 8; e ++) 
	    if (rule[a][b][c][d][e] == -1) {
	      if (sd == ON) rule[a][b][c][d][e] = 8;
	      else rule[a][b][c][d][e] = a;
	    }
}


/* Initialization of cell array */

void load_image(g, cell, x, y, image_file)
     GROUND *g;
     int *cell, x, y;
     char *image_file;
{
  FILE *fp;
  char buf[5000];
  int  w, h, l, data;


  /* Measuring the size of image */

  if ((fp = fopen (image_file, "r")) == NULL) {
    fprintf(stderr, "Image file \"%s\" cannot be opened.\n", image_file);
    exit();
  }

  w = h = 0;

  while (1) {
    if (fgets(buf, 5000, fp) == NULL) break;

    h ++;
    l = strlen(buf);
    if (buf[l - 1] == '\n') l --;
    if (l > w) w = l;
  }

  fclose(fp);


  /* Loading image */

  if ((fp = fopen (image_file, "r")) == NULL) {
    fprintf(stderr, "Image file \"%s\" cannot be opened.\n", image_file);
    exit();
  }

  x = x - w / 2;
  y = y - h / 2;

  if ((x < 0) || (x + w > g->w) || (y < 0) || (y + h > g->h)) {
    fprintf(stderr, "Image file \"%s\" sticks out from the space.\n",
	    image_file);
    exit();
  }

  w = x;
  h = y;

  while (1) {
    
    if ((data = fgetc(fp)) == EOF) break;
    else if (data == '\n') {
      x = w;
      y ++;
    }
    else if ((data == '0') || (data == ' ')) {
      cell[y * g->w + x] = 0;
      x ++;
    }
    else if ((data >= '1') && (data <= '8')) {
      cell[y * g->w + x] = (int) (data - '0');
      x ++;
    }
    else {
      fprintf(stderr, "Illegal state appeared in \"%s\".\n", image_file);
      exit();
    }
  }

  fclose(fp);
}

void init_cell(g, cell, image_file)
     GROUND *g;
     int *cell;
     char *image_file;
{
  int x, y;

  char *image_file_2, *image_file_3, *image_file_4;


  for (x = 0; x < g->w; x ++)
    for (y = 0; y < g->h; y ++) cell[y * g->w + x] = 0;


  image_file_3 = image_file_4 = NULL;

  image_file_2 = strchr(image_file, ',');
  if (image_file_2 != NULL) {
    *image_file_2 = '\0';
    image_file_2 ++;

    image_file_3 = strchr(image_file_2, ',');
    if (image_file_3 != NULL) {
      *image_file_3 = '\0';
      image_file_3 ++;

      image_file_4 = strchr(image_file_3, ',');
      if (image_file_4 != NULL) {
	*image_file_4 = '\0';
	image_file_4 ++;

	if (strchr(image_file_4, ',') != NULL) {
	  fprintf(stderr, "Too many image files.\n");
	  exit();
	}
      }
    }
  }

  if (image_file_2 == NULL) {
    load_image(g, cell, g->w / 2, g->h / 2, image_file);
    return;
  }

  else {
    load_image(g, cell, g->w / 4, g->h / 4, image_file);
    load_image(g, cell, g->w * 3 / 4, g->h * 3 / 4, image_file_2); 
    if (image_file_3 != NULL)
      load_image(g, cell, g->w * 3 / 4, g->h / 4, image_file_3); 
    if (image_file_4 != NULL)
      load_image(g, cell, g->w / 4, g->h * 3 / 4, image_file_4); 
  }
}


/* Drawing subroutines of cell array */

void display_cell(g, x, y, c)
     GROUND *g;
     int x, y, c;
{
  XSetForeground(g->disp, g->gc, g->color[c]);
  XFillRectangle(g->disp, g->window, g->gc,
		 x * g->pixel, y * g->pixel, g->pixel, g->pixel);
}

void display_all_cell(g, cell)
     GROUND *g;
     int *cell;
{
  int x, y;

  XClearWindow(g->disp, g->window);

  for (y = 0; y < g->h; y ++)
    for (x = 0; x < g->w; x ++)
      if (cell[y * g->w + x] != 0) display_cell(g, x, y, cell[y * g->w + x]);

  XFlush(g->disp);
}

void display_time(g, timestep)
     GROUND *g;
     unsigned long timestep;
{
  char buf[30];

  XSetForeground(g->disp, g->gc, g->color[0]);
  XFillRectangle(g->disp, g->window, g->gc, 0, g->h * g->pixel, 100, MARGIN);

  XSetForeground(g->disp, g->gc, g->color[7]);
  sprintf(buf, "Time=%7u", timestep);
  XDrawString(g->disp, g->window, g->gc, 0, g->h * g->pixel + MARGIN, buf, 12);

  XFlush(g->disp);
}

void write_image(g, cell, image_file)
     GROUND *g;
     int *cell;
     char *image_file;
{
  FILE *fp;
  int p, x, y;

  if ((fp = fopen (image_file, "w")) == NULL) {
    fprintf(stderr, "Image file \"%s\" cannot be opened.\n", image_file);
    exit();
  }

  p = 0;
  for (y = 0; y < g->h; y ++) {
    for (x = 0; x < g->w; x ++) {
      if (cell[p] == 0) fprintf(fp, " ");
      else fprintf(fp, "%c", (int)'0' + cell[p]);
      p ++;
    }
    fprintf(fp, "\n");
  }

  fclose(fp);
}


/* Simulation engine */

void simulate_cell(g, cell, ncell, rule)
     GROUND *g;
     int *cell, *ncell, rule[9][9][9][9][9];
{
  int x, y, p, C, T, R, B, L;

  p = 0;

  for (y = 0; y < g->h; y ++)
    for (x = 0; x < g->w; x ++) {
      C = cell[p];

      if (g->boundary == CYCLIC) {
	if (y == 0) T = cell[(g->h - 1) * g->w + x]; else T = cell[p - g->w];
	if (x == g->w - 1) R = cell[p - g->w + 1];   else R = cell[p + 1];
	if (y == g->h - 1) B = cell[x];	             else B = cell[p + g->w];
	if (x == 0) L = cell[p + g->w - 1];          else L = cell[p - 1];
      }

      else {
	if (y == 0)        T = 0; else T = cell[p - g->w];
        if (x == g->w - 1) R = 0; else R = cell[p + 1];
	if (y == g->h - 1) B = 0; else B = cell[p + g->w];
	if (x == 0)        L = 0; else L = cell[p - 1];
      }

      ncell[p] = rule[C][T][R][B][L];
      p ++;
    }

  p = 0;

  for (y = 0; y < g->h; y ++)
    for (x = 0; x < g->w; x ++) {
      if (cell[p] != ncell[p]) {
	cell[p] = ncell[p];
	if (g->screen == ON) display_cell(g, x, y, cell[p]);
      }
      p ++;
    }
}


/* Human-machine interface on X-Window */

void check_event(g, cell, timestep)
     GROUND *g;
     int *cell;
     unsigned long timestep;
{
  static int ctrl = OFF;

  XEvent eve;
  char str[2], buf[256];
  KeySym sym;

  while (1) {

    if (XCheckWindowEvent(g->disp, g->window, ButtonPressMask, &eve) == True) {
      if (eve.xbutton.button == Button1) {
	g->screen ^= 1;
	if (g->screen == ON) display_all_cell(g, cell);
	else {
	  XClearWindow(g->disp, g->window);
	  XFlush(g->disp);
	}
	if (g->time == ON) display_time(g, timestep);
      }
      else if (eve.xbutton.button == Button2) break;
      else if (eve.xbutton.button == Button3) {
	g->pause ^= 1;
	if (g->pause == ON) printf("Time = %u.\n", timestep);
      }
    }

    if (XCheckWindowEvent(g->disp, g->window, ExposureMask, &eve) == True) {
      if (g->screen == ON) display_all_cell(g, cell);
      if (g->time == ON) display_time(g, timestep);
    }

    if (XCheckWindowEvent(g->disp, g->window, KeyReleaseMask, &eve) == True) {
      XLookupString(&(eve.xkey), str, 1, &sym, NULL);
      if ((sym == XK_Control_L) || (sym == XK_Control_R)) ctrl = OFF;
    }

    if (XCheckWindowEvent(g->disp, g->window, KeyPressMask, &eve) == True) {
      XLookupString(&(eve.xkey), str, 1, &sym, NULL);
      if ((sym == XK_Control_L) || (sym == XK_Control_R)) ctrl = ON;

      else if ((ctrl == ON) && (sym == XK_c)) {
	fprintf(stderr, "Abort.\n");
	exit();
      }

      else if ((ctrl == ON) && (sym == XK_s)) {
	fprintf(stderr, "Writing the spatial image (time=%d). File name? ",
		timestep);
	gets(buf);
	if (*buf == '\0') fprintf(stderr, "Cancelled.\n");
	else {
	  write_image(g, cell, buf);
	  fprintf(stderr, "Finished.\n");
	}
      }
    }

    if (g->pause == OFF) break;
  }
}


/* Main */

main(argc, argv, envp)
     int  argc;
     char **argv, **envp;
{
  GROUND grd;
  int *cell, *ncell, rule[9][9][9][9][9];
  int sd, extension, symmetry, i, d;
  char rule_file[256], image_file[256];
  unsigned long interval, timestep, pausecount;


  /* Setting default parameters */

  sd = OFF;
  extension = ON;
  symmetry = ON;

  grd.w = 200;
  grd.h = 200;
  grd.boundary = CYCLIC;

  grd.pixel = 3;
  grd.screen = ON;
  grd.pause = ON;
  grd.time = OFF;

  interval = 0;
  timestep = 0;
  pausecount = 0;

  sprintf(rule_file, "rule_by_langton");
  sprintf(image_file, "image.6");


  /* Reading arguments */

  for (i = 1; i < argc; i ++) {

    if ((strcmp(argv[i], "-help") == 0) || (strcmp(argv[i], "-h") == 0)) {
      printf("Usage:\n");
      printf("loops [-help/-h][-sd][-free][-time]\n");
      printf("      [-begin beginning_time][-interval interval_of_pausing]\n");
      printf("      [-space width_of_space height_of_space]\n");
      printf("      [-pixel size_of_one_cell]\n");
      printf("      [-rule rule_file][-image file1[,file2[,file3[,file4]]]]\n");
      printf("      [-nosymmetry][-noextension]\n");
      exit();
    }

    else if (strcmp(argv[i], "-sd") == 0) sd = ON;
    else if (strcmp(argv[i], "-free") == 0) grd.boundary = FREE;
    else if (strcmp(argv[i], "-time") == 0) grd.time = ON;

    else if (strcmp(argv[i], "-begin") == 0) {
      i ++;
      d = (unsigned long) atoi(argv[i]);
      if ((i>= argc) || (d < 0))
	fprintf(stderr, "Illegal operand of -begin \"%s\".\n", argv[i]);
      else timestep = d;
    }
    else if (strcmp(argv[i], "-interval") == 0) {
      i ++;
      d = (unsigned long) atoi(argv[i]);
      if ((i>= argc) || (d < 0))
	fprintf(stderr, "Illegal operand of -interval \"%s\".\n", argv[i]);
      else interval = pausecount = d;
    }
    else if (strcmp(argv[i], "-space") == 0) {
      i ++;
      d = atoi(argv[i]);
      if ((i>= argc) || (d < 1))
	fprintf(stderr, "Illegal operand of -space (width) \"%s\".\n", argv[i]);
      else grd.w = d;
      i ++;
      d = atoi(argv[i]);
      if ((i>= argc) || (d < 1))
	fprintf(stderr, "Illegal operand of -space (height) \"%s\".\n", argv[i]);
      else grd.h = d;
    }
    else if (strcmp(argv[i], "-pixel") == 0) {
      i ++;
      d = atoi(argv[i]);
      if ((i>= argc) || (d < 1))
	fprintf(stderr, "Illegal operand of -pixel \"%s\".\n", argv[i]);
      else grd.pixel = d;
    }
    else if (strcmp(argv[i], "-rule") == 0) {
      i ++;
      if (i>= argc)
	fprintf(stderr, "Illegal operand of -rule \"%s\".\n", argv[i]);
      else strcpy(rule_file, argv[i]);
    }
    else if (strcmp(argv[i], "-image") == 0) {
      i ++;
      if (i>= argc)
	fprintf(stderr, "Illegal operand of -image \"%s\".\n", argv[i]);
      else strcpy(image_file, argv[i]);
    }

    else if (strcmp(argv[i], "-nosymmetry") == 0) symmetry = OFF;
    else if (strcmp(argv[i], "-noextension") == 0) extension = OFF;

    else fprintf(stderr, "Illegal option \"%s\".\n", argv[i]);
  }

  /* Initialization */

  cell  = (int *) malloc(sizeof(int) * grd.w * grd.h);
  ncell = (int *) malloc(sizeof(int) * grd.w * grd.h);

  load_rule(rule, rule_file, symmetry);
  set_undefined_rule(rule, extension, sd);
  init_cell(&grd, cell, image_file);

  open_window(&grd);
  set_color(&grd);
  display_all_cell(&grd, cell);


  /* Simulation */

  if (grd.time == ON) display_time(&grd, timestep);

  printf("Time = %u.\n", timestep);
  check_event(&grd, cell, timestep);

  while (1) {
    simulate_cell(&grd, cell, ncell, rule);
    if (grd.screen == ON) XFlush(grd.disp);

    timestep ++;
    if (grd.time == ON) display_time(&grd, timestep);
    if (interval != 0) {
      pausecount --;
      if (pausecount == 0) {
	pausecount = interval;
	grd.pause = ON;
      }
    }
    if (grd.pause == ON) printf("Time = %u.\n", timestep);
    check_event(&grd, cell, timestep);
  }
}
